# -*- coding: utf-8 -*-

import logging

import constants
import os
import utils


class ServiceState(object):
    def __init__(self, pid_file_path=constants.DEFAULT_PID_FILE_LOCATION):
        self.job_worker_pids = []
        self.retry_worker_pids = []
        self.dlq_worker_pids = []
        self.pid_file = pid_file_path
        if os.path.exists(self.pid_file):
            tmp_obj = utils.load_obj(self.pid_file)
            self.__dict__ = tmp_obj.__dict__
        else:
            self.persist_state()

    def persist_state(self):
        """
        Persist the library pid state to the pid file
        :return: 
        """
        utils.save_obj(self, self.pid_file)

    def get_pid_list(self, type):
        """
        Get the pid list for the give type of worker
        :param type: type of worker
        :return: 
        """
        if type == constants.WORK_QUEUE:
            return self.job_worker_pids
        elif type == constants.RETRY_QUEUE:
            return self.retry_worker_pids
        elif type == constants.DEAD_LETTER_QUEUE:
            return self.dlq_worker_pids
        else:
            raise KeyError("Invalid key : " + str(type))

    def add_worker_pid(self, type, pid):
        """
        add a pid for the worker
        :param type: type of worker
        :param pid: pid to be added
        :return: 
        """
        if type == constants.WORK_QUEUE:
            self.job_worker_pids.append(pid)
        elif type == constants.RETRY_QUEUE:
            self.retry_worker_pids.append(pid)
        elif type == constants.DEAD_LETTER_QUEUE:
            self.dlq_worker_pids.append(pid)
        else:
            raise KeyError("Invalid key : " + str(type))
        self.persist_state()

    def remove_worker_pid(self, worker_type, pid):
        """
        remove a worker pid from the list
        :param worker_type: type of worker
        :param pid: pid to be removed
        :return: 
        """
        try:
            if worker_type == constants.WORK_QUEUE:
                self.job_worker_pids.remove(pid)
            elif worker_type == constants.RETRY_QUEUE:
                self.retry_worker_pids.remove(pid)
            elif worker_type == constants.DEAD_LETTER_QUEUE:
                self.dlq_worker_pids.remove(pid)
            self.persist_state()
        except ValueError as e:
            logger = logging.getLogger(self.__class__.__name__)
            logger.warning("unable to remove {} as not found in {} pid list with exception: {}."
                           .format(pid, worker_type, e.message))

    def refresh_workers_pid(self, worker_type):
        """
        refresh the state of pids
        :param worker_type: the type of worker
        :return: 
        """
        pid_list = list(self.get_pid_list(worker_type))
        # job worker
        for pid in pid_list:
            if not utils.is_process_running(pid):
                self.remove_worker_pid(worker_type, pid)

    def refresh_all_workers_pid(self):
        """
        refresh all the worker queue pids
        :return: 
        """
        self.refresh_workers_pid(constants.WORK_QUEUE)
        self.refresh_workers_pid(constants.RETRY_QUEUE)
        self.refresh_workers_pid(constants.DEAD_LETTER_QUEUE)



# -*- coding: utf-8 -*-

import json
import random
import string
import time

import constants
from easy_api import EasyApi
from easyjoblite.response import EasyResponse
from exception import UnableToCreateJob


class EasyJob(object):
    def __init__(self):
        self.id = "1"
        self.type = None
        self.tag = "unknown"
        self.job_api = None
        self.no_of_retries = 0
        self.data = None
        self.notification_handler = None
        self.errors = []
        self.response = None

    @property
    def api(self):
        return self.job_api.api

    def should_notify_error(self):
        return self.notification_handler is not None

    @classmethod
    def generate_id(cls):
        return "-".join(["".join(str(time.time()).split('.')), random.choice(string.digits)])

    @classmethod
    def create(cls, api, job_type, remote_call_type=None, data=None, api_request_headers=None,
               content_type=None, tag=None, notification_handler=None):
        job = cls()
        job.type = job_type
        job.id = cls.generate_id()
        job.job_api = EasyApi.create(job_type, api, api_request_headers, remote_call_type, content_type)
        if notification_handler:
            job.notification_handler = EasyApi.create(job_type,
                                                      notification_handler,
                                                      api_request_headers,
                                                      "post",
                                                      content_type
                                                      )
        if data:
            job.data = json.dumps(data)
        if tag:
            job.tag = tag
        return job

    @classmethod
    def create_from_dict(cls, dict_data):
        try:
            job = cls()
            if "id" in dict_data:
                job.id = dict_data["id"]
            if "type" in dict_data:
                job.type = dict_data["type"]
            if "tag" in dict_data:
                job.tag = dict_data["tag"]
            job.data = dict_data.get("data", None)
            if "job_api" in dict_data:
                job.job_api = EasyApi.create_from_dict(dict_data["job_api"])
            job.no_of_retries = dict_data.get("no_of_retries", 0)
            if "errors" in dict_data:
                job.errors = dict_data["errors"]
            if "notification_handler" in dict_data:
                job.notification_handler = EasyApi.create_from_dict(dict_data["notification_handler"])
            if "response" in dict_data:
                job.response = EasyResponse(dict_data["response"])
        except Exception as e:
            raise UnableToCreateJob(e.message, dict_data)
        return job

    @classmethod
    def create_dummy_clone_from_dict(cls, dict_data):
        try:
            job = cls()
            if "id" in dict_data:
                job.id = dict_data["id"]
            if "type" in dict_data:
                job.type = dict_data["type"]
            if "tag" in dict_data:
                job.tag = dict_data["tag"]
        except Exception as e:
            raise UnableToCreateJob(e.message, dict_data)
        return job

    def to_dict(self):
        dict_data = {
            "id": self.id,
            "type": self.type,
            "tag": self.tag,
            "data": self.data,
            "no_of_retries": self.no_of_retries,
            "errors": self.errors
        }

        if self.response:
            dict_data["response"] = self.response.__dict__
        else:
            dict_data["response"] = None

        if self.job_api:
            dict_data["job_api"] = self.job_api.to_dict()
        else:
            dict_data["job_api"] = None

        if self.notification_handler:
            dict_data["notification_handler"] = self.notification_handler.to_dict()
        else:
            dict_data["notification_handler"] = None
        return dict_data

    def add_error(self, error):
        self.errors.append(error)

    def get_response(self):
        return self.response.__dict__

    def execute(self, data, async_timeout=constants.DEFAULT_ASYNC_TIMEOUT):
        api_dict = dict(job_id=self.id, tag=self.tag, data=data)
        ret_val = self.job_api.execute(api_dict, async_timeout)

        if ret_val.status_code != 200:
            self.errors.append(ret_val.__dict__)
        else:
            self.response = ret_val

        return ret_val

    def notify_error(self, data, async_timeout=constants.DEFAULT_ASYNC_TIMEOUT):
        if self.should_notify_error():
            error_data = dict(job_id=self.id, tag=self.tag, data=data, errors=self.errors)
            if self.job_api:
                error_data["api"] = self.job_api.api
            return self.notification_handler.execute(error_data, async_timeout)
        else:
            return EasyResponse(400, "No notification handler defined.")

    def increment_retries(self):
        self.no_of_retries += 1



# -*- coding: utf-8 -*-

import requests

# Job types
JOB_TYPE = 'job_type'
API_LOCAL = 'local'
API_REMOTE = 'remote'
job_call_types = {API_REMOTE, API_LOCAL}

# Remote call type
REMOTE_CALL_TYPE = 'remote_call_type'
remote_call_type = {
    'post': requests.post,
    'get': requests.get,
    'put': requests.put,
    'delete': requests.delete
}

# Errors to be notified
errors_to_be_notified = {
    'booking.confirm': {400, 410},
    'booking.cancel': {400}
}

# Queue Types
EXCHANGE = "exchange"
DEAD_LETTER_QUEUE = "dead"
RETRY_QUEUE = "retry"
WORK_QUEUE = "work"
BUFFER_QUEUE = "buffer"

# rabbit mq exchange and queue prefix
rabbit_mq_prefix = {
    EXCHANGE: "-ejl-exchange",
    WORK_QUEUE: "-work-queue",
    DEAD_LETTER_QUEUE: "-dead-letter-queue",
    RETRY_QUEUE: "-retry-queue",
    BUFFER_QUEUE: "-buffer-queue",
}

# Default configs
DEFAULT_APP_ID = "DefaultApp"
DEFAULT_RMQ_URL = 'amqp://guest:guest@localhost:5672//'
DEFAULT_ASYNC_TIMEOUT = 120
DEFAULT_MAX_JOB_RETRIES = 3
DEFAULT_ERROR_Q_CON_SLEEP_DURATION = 15
DEFAULT_MAX_WORKER_COUNT = 10
DEFAULT_WORKER_COUNT = 3
DEFAULT_RETRY_CONSUMER_COUNT = 1
DEFAULT_DLQ_CONSUMER_COUNT = 3
DEFAULT_IMPORT_PATHS = '.'
DEFAULT_PID_FILE_LOCATION = "/var/tmp/easyjoblite.pid"
DEFAULT_LOG_FILE_PATH = "/var/tmp/easyjoblite.log"
DEFAULT_DL_LOG_FILE = "/var/tmp/easyjoblite_dl.log"
DEFAULT_CONFIG_FILE = "/var/tmp/easyjoblite.yaml"
DEFAULT_HEALTH_CHECK_INTERVAL = 2

BASE_COMMAND = "nohup easyjoblite"
STOP_TYPE_ALL = "all"



# -*- coding: utf-8 -*-

import logging
import os
import pickle
import signal
import sys
import time
import threading
import traceback

from kombu.entity import PERSISTENT_DELIVERY_MODE

from easyjoblite import state
from easyjoblite import constants


def as_text(v):
    if v is None:
        return None
    elif isinstance(v, bytes):
        return v.decode('utf-8')
    elif isinstance(v, str):
        return v
    else:
        raise ValueError('Unknown type %r' % type(v))


def is_string_type(value):
    string_types = (str, unicode)
    if isinstance(value, string_types):
        return True
    else:
        return False


def save_obj(obj, path):
    """
    save a python object to a file
    :param obj: the object to be saved
    :param path: the path of the file where to be saved
    :return: None
    """
    with open(path, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(path):
    """
    return the python object saved in the given path
    :param path: the path to be loaded
    :return: 
    """
    logger = logging.getLogger("load_obj")
    retry_count = 3
    while retry_count > 0:
        try:
            with open(path, 'rb') as f:
                return pickle.load(f)
        except (IOError, EOFError) as e:
            logger.warning("failed loading obj with error: {}. Will retry for {} time.".format(retry_count, e.message))
            time.sleep(2)
            retry_count -= 1


def is_process_running(pid):
    """Check whether pid exists in the current process table.
    UNIX only.
    :param pid: the processes id to check
    :return: TRUE/FALSE
    """
    if pid < 0:
        return False
    try:
        ret = os.popen("ps -eo pid,stat | grep {} |  grep -v \'Z\' | awk \'{{ print $1}}\'".format(pid)).read()

        if ret and int(ret) == pid:
            return True
        else:
            return False
    except Exception as e:
        traceback.print_exc()
        logger = logging.getLogger("is_process_running")
        logger.warning("something broke while getting process state so taking it as not running.")
    else:
        return True


def kill_process(pid):
    """
    kill a process for a given pid
    :param pid: pid to be killed
    :return: True if killed sucessfully or raises an exception
    """
    try:
        os.kill(int(pid), signal.SIGTERM)
        # Check if the process that we killed is alive.
    except OSError as ex:
        return True


def get_pid_state_string(pid_list):
    """
    state of the pids in string
    :param pid_list: list of pid's
    :return: 
    """
    total_state = ""

    for pid in pid_list:
        state_str = "STOPPED"
        if is_process_running(pid):
            state_str = "RUNNING"

        total_state += "{}:{}   ".format(pid, state_str)

    if not total_state:
        return "NONE"
    return total_state


def is_main_thread():
    """
    check if current thread is main thread or not
    :return: 
    """
    return threading.current_thread().__class__.__name__ == '_MainThread'


def kill_workers(service_state, type):
    """
    function to kill all the workers of the given type
    :param service_state: current state of the service
    :param type: the type of the worker to kill
    :return: 
    """
    logger = logging.getLogger("kill_workers")
    logger.info("Started killing : " + type + " with list " + str(service_state.get_pid_list(type)))
    pid_list = list(service_state.get_pid_list(type))
    for pid in pid_list:
        kill_process(pid)
        logging.info("Done killing : " + str(pid))


def stop_all_workers(worker_type):
    """
    stops all the workers of the given type
    :param worker_type: 
    :return: 
    """
    logger = logging.getLogger("stop_all_workers")
    service_state = state.ServiceState()
    worker_type_list = [constants.WORK_QUEUE, constants.RETRY_QUEUE, constants.DEAD_LETTER_QUEUE]

    if worker_type in worker_type_list:
        kill_workers(service_state, worker_type)
        logger.info("Done stopping all the workers of worker_type {}".format(worker_type))
    elif worker_type == constants.STOP_TYPE_ALL:
        for local_type in worker_type_list:
            kill_workers(service_state, local_type)
        logger.info("Done stopping all the workers ")
    else:
        raise KeyError
    service_state.refresh_all_workers_pid()


def update_import_paths(import_paths):
    """
    update the import paths in the system
    :param import_paths: 
    :return: 
    """
    if import_paths:
        sys.path = import_paths.split(':') + sys.path


def enqueue(producer, queue_type, job, body):
    """
    enque a job in the given queue
    :param producer: the producer to be used
    :param queue_type: type of queue (worker, retry, dead)
    :param job: the job object
    :param body: the body payload of the job
    :return: none
    """
    routing_key = "{type}.{tag}".format(type=queue_type, tag=job.tag)
    headers = job.to_dict()
    producer.publish(body=body,
                     headers=headers,
                     routing_key=routing_key,
                     delivery_mode=PERSISTENT_DELIVERY_MODE)
